@@ -77,8 +77,9 @@ read_UCHAR(SerdReader* reader, Ref dest, uint32_t* char_code)
 		}
 	}
 
-	uint32_t code;
-	sscanf((const char*)buf, "%X", &code);
+	char*          endptr = NULL;
+	const uint32_t code   = strtoul((const char*)buf, &endptr, 16);
+	assert(endptr == (char*)buf + length);
 
 	unsigned size = 0;
 	if (code < 0x00000080) {
@@ -105,14 +106,17 @@ read_UCHAR(SerdReader* reader, Ref dest, uint32_t* char_code)
 		buf[3] = 0x80 | (uint8_t)(c & 0x3F);
 		c >>= 6;
 		c |= (16 << 12);  // set bit 4
+        // fallthru
 	case 3:
 		buf[2] = 0x80 | (uint8_t)(c & 0x3F);
 		c >>= 6;
 		c |= (32 << 6);  // set bit 5
+        // fallthru
 	case 2:
 		buf[1] = 0x80 | (uint8_t)(c & 0x3F);
 		c >>= 6;
 		c |= 0xC0;  // set bits 6 and 7
+        // fallthru
 	case 1:
 		buf[0] = (uint8_t)c;
 	}
@@ -209,7 +213,7 @@ static SerdStatus
 read_utf8_code(SerdReader* reader, Ref dest, uint32_t* code, uint8_t c)
 {
 	uint32_t   size;
-	uint8_t    bytes[4];
+	uint8_t    bytes[4] = { 0, 0, 0, 0 };
 	SerdStatus st = read_utf8_bytes(reader, bytes, &size, c);
 	if (st) {
 		push_bytes(reader, dest, replacement_char, 3);
@@ -301,32 +305,28 @@ read_STRING_LITERAL_LONG(SerdReader* reader, SerdNodeFlags* flags, uint8_t q)
 	Ref ref = push_node(reader, SERD_LITERAL, "", 0);
 	while (!reader->status) {
 		const uint8_t c = peek_byte(reader);
-		uint32_t      code;
-		switch (c) {
-		case '\\':
+		if (c == '\\') {
 			eat_byte_safe(reader, c);
+			uint32_t code;
 			if (!read_ECHAR(reader, ref, flags) &&
 			    !read_UCHAR(reader, ref, &code)) {
 				r_err(reader, SERD_ERR_BAD_SYNTAX,
 				      "invalid escape `\\%c'\n", peek_byte(reader));
 				return pop_node(reader, ref);
 			}
-			break;
-		default:
-			if (c == q) {
-				eat_byte_safe(reader, q);
-				const uint8_t q2 = eat_byte_safe(reader, peek_byte(reader));
-				const uint8_t q3 = peek_byte(reader);
-				if (q2 == q && q3 == q) {  // End of string
-					eat_byte_safe(reader, q3);
-					return ref;
-				}
-				*flags |= SERD_HAS_QUOTE;
-				push_byte(reader, ref, c);
-				read_character(reader, ref, flags, q2);
-			} else {
-				read_character(reader, ref, flags, eat_byte_safe(reader, c));
+		} else if (c == q) {
+			eat_byte_safe(reader, q);
+			const uint8_t q2 = eat_byte_safe(reader, peek_byte(reader));
+			const uint8_t q3 = peek_byte(reader);
+			if (q2 == q && q3 == q) {  // End of string
+				eat_byte_safe(reader, q3);
+				return ref;
 			}
+			*flags |= SERD_HAS_QUOTE;
+			push_byte(reader, ref, c);
+			read_character(reader, ref, flags, q2);
+		} else {
+			read_character(reader, ref, flags, eat_byte_safe(reader, c));
 		}
 	}
 	return ref;
@@ -646,10 +646,12 @@ read_IRIREF(SerdReader* reader)
 			if (c <= 0x20) {
 				if (isprint(c)) {
 					r_err(reader, SERD_ERR_BAD_SYNTAX,
-					      "invalid IRI character `%c' (escape %%%02X)\n", c, c);
+					      "invalid IRI character `%c' (escape %%%02X)\n",
+					      c, (unsigned)c);
 				} else {
 					r_err(reader, SERD_ERR_BAD_SYNTAX,
-					      "invalid IRI character (escape %%%02X)\n", c, c);
+					      "invalid IRI character (escape %%%02X)\n",
+					      (unsigned)c);
 				}
 				if (reader->strict) {
 					return pop_node(reader, ref);
@@ -801,18 +803,6 @@ except:
 	return r_err(reader, SERD_ERR_BAD_SYNTAX, "bad literal syntax\n");
 }
 
-inline static bool
-is_token_end(uint8_t c)
-{
-	switch (c) {
-	case 0x9: case 0xA: case 0xD: case 0x20: case '\0':
-	case '#': case '.': case ';': case '<':
-		return true;
-	default:
-		return false;
-	}
-}
-
 static bool
 read_verb(SerdReader* reader, Ref* dest)
 {
@@ -827,8 +817,9 @@ read_verb(SerdReader* reader, Ref* dest)
 	const SerdStatus st      = read_PN_PREFIX(reader, *dest);
 	bool             ate_dot = false;
 	SerdNode*        node    = deref(reader, *dest);
+	const uint8_t    next    = peek_byte(reader);
 	if (!st && node->n_bytes == 1 && node->buf[0] == 'a' &&
-	    is_token_end(peek_byte(reader))) {
+	    next != ':' && !is_PN_CHARS_BASE(next)) {
 		pop_node(reader, *dest);
 		return (*dest = push_node(reader, SERD_URI, NS_RDF "type", 47));
 	} else if (st > SERD_FAILURE ||
@@ -874,7 +865,7 @@ read_BLANK_NODE_LABEL(SerdReader* reader, bool* ate_dot)
 		*ate_dot = true;
 	}
 
-	if (reader->syntax == SERD_TURTLE) {
+	if (fancy_syntax(reader)) {
 		if (is_digit(n->buf[reader->bprefix_len + 1])) {
 			if ((n->buf[reader->bprefix_len]) == 'b') {
 				((char*)n->buf)[reader->bprefix_len] = 'B';  // Prevent clash
@@ -1187,7 +1178,7 @@ read_subject(SerdReader* reader, ReadContext ctx, Ref* dest, char* s_type)
 }
 
 static Ref
-read_labelOrSubject(SerdReader* reader, ReadContext ctx)
+read_labelOrSubject(SerdReader* reader)
 {
 	Ref  subject = 0;
 	bool ate_dot = false;
@@ -1314,11 +1305,11 @@ read_directive(SerdReader* reader)
 static bool
 read_wrappedGraph(SerdReader* reader, ReadContext* ctx)
 {
-	bool ate_dot = false;
-	char s_type  = 0;
 	TRY_RET(eat_byte_check(reader, '{'));
 	read_ws_star(reader);
 	while (peek_byte(reader) != '}') {
+		bool ate_dot = false;
+		char s_type  = 0;
 		ctx->subject = 0;
 		Ref subj = read_subject(reader, *ctx, &ctx->subject, &s_type);
 		if (!subj && ctx->subject) {
@@ -1388,7 +1379,7 @@ read_n3_statement(SerdReader* reader)
 			ret = read_prefixID(reader, true, false);
 		} else if (!tokcmp(reader, ctx.subject, "graph", 5)) {
 			read_ws_star(reader);
-			TRY_RET((ctx.graph = read_labelOrSubject(reader, ctx)));
+			TRY_RET((ctx.graph = read_labelOrSubject(reader)));
 			read_ws_star(reader);
 			TRY_RET(read_wrappedGraph(reader, &ctx));
 			pop_node(reader, ctx.graph);
@@ -1407,7 +1398,10 @@ read_n3_statement(SerdReader* reader)
 		} else if (!subj) {
 			ret = r_err(reader, SERD_ERR_BAD_SYNTAX, "bad subject\n");
 		} else if (!read_triples(reader, ctx, &ate_dot)) {
-			ret = (s_type == '[');
+			if (!(ret = (s_type == '[')) && ate_dot) {
+				ret = r_err(reader, SERD_ERR_BAD_SYNTAX,
+				            "unexpected end of statement\n");
+			}
 		} else if (!ate_dot) {
 			read_ws_star(reader);
 			ret = (eat_byte_check(reader, '.') == '.');
